"""
The `article.py` module handles fetching and parsing individual articles.
The articles are extracted by url, with the url pointing to an article.
"""

import json
import re
import requests
from lxml.html import HtmlElement, fromstring
from typing import Dict, Optional


class Article:
    """A class used to represent a specific article.

    Attributes
    ----------
    url : str
        The URL of the article.
    doc : Optional[HtmlElement]
        The parsed HTML document of the article.

    Methods
    -------
    fetch() -> None:
        Fetches the article page and sets the `doc` attribute with the parsed HTML document.
    parse() -> dict[str, Any]:
        Parses the fetched article HTML to extract necessary details.
    """

    def __init__(self, url: str):
        """Constructs an Article object.

        Parameters
        ----------
        url : str
            The URL of the article.
        """
        self.url: str = url
        self.doc: Optional[HtmlElement] = None

    def fetch(self):
        """Fetches the article HTML content and sets the `doc` attribute with the parsed HTML document.

        Raises
        ------
        requests.exceptions.RequestException
            If the HTTP request to fetch the article page fails.
        """
        resp: requests.Response = requests.get(self.url)
        self.doc: HtmlElement = fromstring(resp.text)

    def parse(self) -> dict:
        """Parses the fetched article HTML to extract necessary details.

        Returns
        -------
        dict[str, Any]
            A dictionary containing article details such as URL, ID, channel, subchannel, headline,
            intro, text, topics, author, comments_enabled, dates (created, modified, published), and breadcrumbs.

        Raises
        ------
        AssertionError
            If `fetch` method was not called before parsing.
        """
        assert self.doc is not None, "Fetch the document first before parsing."

        linked_data_content: str = self.doc.xpath(
            '//script[@type="application/ld+json"]/text()'
        )[0]
        linked_data: dict = json.loads(linked_data_content)
        linked_data_by_type: dict[str, dict] = {
            ld_entry["@type"]: ld_entry for ld_entry in linked_data
        }
        news_linked_data: dict = linked_data_by_type["NewsArticle"]

        settings: dict = json.loads(
            self.doc.xpath('//script[@type="application/settings+json"]/text()')[0]
        )
        info: dict = settings["editorial"]["info"]

        text_nodes: list[HtmlElement] = self.doc.cssselect(
            "main .word-wrap > p, main .word-wrap > h3, main .word-wrap > ul > li, main .word-wrap > ol > li"
        )
        full_text: str = "\n".join([node.text_content() for node in text_nodes])
        text: str = re.sub(r"\n+", "\n", full_text).strip()

        article_information: dict = {
            "url": self.doc.xpath('string(//link[@rel="canonical"]/@href)'),
            "id": info["article_id"],
            "channel": info["channel"],
            "subchannel": info["subchannel"],
            "headline": {"main": info["headline"], "social": info["headline_social"]},
            "intro": info["intro"],
            "text": text,
            "topics": info["topics"],
            "author": settings["editorial"]["author"],
            "comments_enabled": settings["editorial"]["attributes"][
                "is_comments_enabled"
            ],
            "date_created": news_linked_data["dateCreated"],
            "date_modified": news_linked_data["dateModified"],
            "date_published": news_linked_data["datePublished"],
            "breadcrumbs": [
                breadcrumb["item"]["name"]
                for breadcrumb in linked_data_by_type["BreadcrumbList"][
                    "itemListElement"
                ]
            ],
        }

        return article_information
